"""
GameBox - A beginner-friendly Python 2D game development library.
--------------------------------------------------------------
GameBox makes it easy to build 2D games with graphics, sound, and UI in just a few lines of code.
"""


__version__ = "0.3.0"
__author__ = "Sam Fertig"

#____imports____
from ._game import Game
from .basics.cammera import Cammera
from .basics._shapes import Rect
from .player._player import Player
from .basics.utils import clamp, moveTward, zeroOut
from .tilemap._tilemap import TileMap
from.helpers._input import Keys
from .ui._basicUI import Image

from .GameLevel_ui._sprites import Sprite_2d, Animated_Sprite2D, AnimationPlayer2D, split_image
from.helpers._Conditions import Conditions


__all__ = [
    "Game",
    "Cammera",
    "Rect",
    "Player",
    "clamp",
    "moveTward",
    "zeroOut",
    "TileMap",
    "Keys",
    "Image",
    "Sprite_2d",
    "Animated_Sprite2D",
    "AnimationPlayer2D",
    "split_image",
    "Conditions",
]




import pygame
import math
import numpy as np

def clamp(value, min_value, max_value):
    return max(min(value, max_value), min_value)
def moveTward(value, target, speed):
    if value < target:
        value += speed
    elif value > target:
        value -= speed
    return value

def zeroOut(value, max):
    if value < 0 and value > -max: value = 0
    if value > 0 and value < max: value = 0
    return value



import pygame
import numpy as np

from..helpers._input import Keys
from ..basics._net import Global
from ._collisionDef import _tileCollisionDefs

from._editorBrushes import _brushPencil, _collisionsPencil

class _tilemapEditor:
    def __init__(self, tilemap, activation):
        self.tilemap = tilemap
        self.activation = activation
        self.active = False

        self.mode = _brushPencil()

        self.changes = {
            "pencil": Keys.b,
            "collisions": Keys.c
            
        }

    def _update(self):
        if  self.active:
            self.change_mode()
            #editor stuff
            self.mode.update(self.tilemap)
        #toggle
        if Keys.is_pressed(self.activation): self.active = not self.active

    def change_mode(self):
        #key presses
        if Keys.is_pressed(self.changes["pencil"]): self.mode = _brushPencil()
        elif Keys.is_pressed(self.changes["collisions"]): self.mode = _collisionsPencil(self.tilemap)
    


import pygame
import os
from src.GameBox import *

width, height = 800, 600

game = Game(width, height, "blue", "First Game!")
screen = game.get_screen()

cam = Cammera()

Keys.init()

player = Player((width / 2, height / 2), (64, 64), "green", False)
player.add_physics(1.0, 3.0, 16, 7.0, 0.5)

player.sprite.add_animation_player(2.0)
player.sprite.add_animation("Walk", "tests/Player.png", (5, 1), (32, 32), 5, 0.075, 2.0, 1)
player.sprite.add_animation("Idle", "tests/Player.png", (5, 1), (32, 32), 1, 0.075, 2.0, 1)

player.sprite.set_animation("Idle")


stateTree = {
    "Idle": {Conditions.velocity_any: "Walk"},
    "Walk": {Conditions.velocity_none: "Idle"}
}
player.sprite.set_states(stateTree, "Walk")

rect = Rect((width / 2, height / 4), (64, 64), "red", True)


cam.set_follow_target(player)
running = True
while running:
    events = pygame.event.get()
    for event in events:
        if event.type == pygame.QUIT:
            running = False
    
    player.top_down_movement()
    game.update(events, 60)

    #print player state
    os.system("cls")
    print(player.sprite.state)

    if Keys.is_pressed(pygame.K_i):
        game._set_debug(not game._get_debug_state())

    
game.quit()
pygame.quit()
os.system("cls")



import pygame

class _tileCollisionDefs:
    def __init__(self, tileDim):
        width, height = tileDim


        self.rects = ["full", "none", "halfLeft", "halfRight", "halfTop", "halfBottom", "topRight", 
                      "topLeft", "bottomRight", "bottomLeft", "dot"]
        
        self.num = len(self.rects) + 1

        self.full = pygame.Rect(0, 0, width, height)

        self.halfLeft = pygame.Rect(0, 0, width / 2, height)
        self.halfRight = pygame.Rect(width / 2, 0, width / 2, height)
        self.halfTop = pygame.Rect(0, 0, width, height / 2)
        self.halfBottom = pygame.Rect(0, height / 2, width, height / 2)

        self.topRight = pygame.Rect(width / 2, 0, width / 2, height / 2)
        self.topLeft = pygame.Rect(0, 0, width / 2, height / 2)
        self.bottomRight = pygame.Rect(width / 2, height / 2, width / 2, height / 2)
        self.bottomLeft = pygame.Rect(0, height / 2, width / 2, height / 2)

        self.dot = pygame.Rect(0, 0, 3, 3)

        self.center = pygame.Rect(width / 2, height / 2, width / 2, height / 2)

        self.none = pygame.Rect(0, 0, 0, 0)

    def regenerate(self, tileDim):
        width, height = tileDim
        self.full = pygame.Rect(0, 0, width, height)

        self.halfLeft = pygame.Rect(0, 0, width / 2, height)
        self.halfRight = pygame.Rect(width / 2, 0, width / 2, height)
        self.halfTop = pygame.Rect(0, 0, width, height / 2)
        self.halfBottom = pygame.Rect(0, height / 2, width, height / 2)

        self.center = pygame.Rect(width / 2, height / 2, width / 2, height / 2)




from src.GameBox import *
import pygame
import os

width, height = 1400, 900
game = Game(width, height, "blue", "First Game!")
win = game.get_screen()

Keys.init()

cam = Cammera()

player = Player((width / 2, height / 4), (64, 64), "green", False)  
player.add_physics(1.0, 3.0, 16, 7.0, 0.5)

map = TileMap("tests/levelTiles.png", (16, 16), 5, (25, 25), 0)
map.load_map_from_json("tests/testMap.json")
map.activate_editor(Keys.tab)

player.sprite.add_animated_sprite_2d("tests/Player.png", (5, 1), (32, 32), 5, 0.075, 2.0, True, 1)

#player.set_dim_as_sprite()

cam.set_follow_target(player)
#cam.move(-50, -50)

running = True
while running:
    events = pygame.event.get()
    for event in events:
        if event.type == pygame.QUIT:
            running = False

    player.top_down_movement()


    game.update(events, 60)


game.quit()
pygame.quit()
os.system("cls")


import pygame
import numpy as np

from ..basics._net import Global

class Animation:
    def __init__(self, image, tileDim, startPos, frames, dur):
        if type(image) == str:
            image = pygame.image.load(image)

        #others
        self.dur = dur
        self.currentFrame = 0
        self.currentDur = 0

        #get frames from image
        self.frames = []
        x, y = startPos
        x *= tileDim[0]
        y *= tileDim[1]
        for i in range(frames):
            self.frames.append(image.subsurface(x, y, tileDim[0], tileDim[1]))
            x += tileDim[0]
            if x >= image.get_width():
                x = 0
                y += tileDim[1]

    def update(self, dt):
        self.currentDur += dt
        if self.currentDur >= self.dur:
            self.currentDur = 0
            self.currentFrame += 1
            if self.currentFrame >= len(self.frames):
                self.currentFrame = 0

    def getFrame(self):
        return self.frames[self.currentFrame]





import pygame
import numpy as np

from ..basics._net import Global
from ..basics.utils import clamp, moveTward, zeroOut

class _playerControler:
    def __init__(self, player: object, speed: float = 1.0, gravity: float = 0.0, jump: float = 0.0, maxV: float = 10.0, airRes: float = 0.2):
        self.speed = speed
        self.gravity = gravity
        self.jump = jump
        self.player = player
        self.maxV = maxV
        self.airRes = airRes

    def update(self):
        self.player.vx = clamp(self.player.vx, -self.maxV, self.maxV)
        #self.player.vy = clamp(self.player.vy, -self.maxV, self.maxV)

        self.player.vx = moveTward(self.player.vx, 0, self.airRes)
        self.player.vy = moveTward(self.player.vy, 0, self.airRes)

        self.player.vx = zeroOut(self.player.vx, 0.3)
        self.player.vy = zeroOut(self.player.vy, 0.3)

        if self.player.gravity and not self.player.onGround: self.player.vy += self.gravity




    def move(self, x, y):
        self.player.vx += x * self.speed
        self.player.vy += y * self.speed

    def top_down_movement(self):
        keys = pygame.key.get_pressed()
        if keys[pygame.K_a]:
            self.move(-1, 0)
        if keys[pygame.K_d]:
            self.move(1, 0)
        if keys[pygame.K_w]:
            self.move(0, -1)
        if keys[pygame.K_s]:
            self.move(0, 1)

    def platforming_movement(self):
        keys = pygame.key.get_pressed()
        if keys[pygame.K_a]:
            self.move(-1, 0)
        if keys[pygame.K_d]:
            self.move(1, 0)
        if keys[pygame.K_SPACE] and self.player.onGround:
            self.player.vy = -self.jump




import pygame
import numpy as np

from ..basics._net import Global
from ..helpers._input import Keys
from ._collisionDef import _tileCollisionDefs

class _brushPencil():
    def __init__(self):
        self.selectedTile = 1
        self.mode = "paint"

    def update(self, tilemap):
        #get all mouse calculations
        x, y = Keys.mouse_x, Keys.mouse_y
        x += Global.cam.x
        y += Global.cam.y
        mx = x // tilemap.tileDim[0] * tilemap.tileDim[0]
        my = y // tilemap.tileDim[1] * tilemap.tileDim[1]

        #get mode
        x, y = Keys.mouse_x, Keys.mouse_y
        if x > tilemap.tileset.get_size()[0] * tilemap.tilescale / 2 or y > tilemap.tileset.get_size()[1] * tilemap.tilescale / 2:
            self.mode = "paint"
        else:
            self.mode = "select"

        #--show tileset
        tile = tilemap.tiles[self.selectedTile]
        image = pygame.transform.scale_by(tilemap.tileset, tilemap.tilescale / 2)
        Global.screen.blit(image, (0, 0))
        #--show outlined sellected tile
        x, y = tilemap.tilePosInImage[self.selectedTile]
        x *= tilemap.tilescale / 2
        y *= tilemap.tilescale / 2
        width = tilemap.orginDim[0] * tilemap.tilescale / 2
        height = tilemap.orginDim[1] * tilemap.tilescale / 2
        outline = pygame.Rect(x, y, width, height)
        pygame.draw.rect(Global.screen, "white", outline, 2)
        #other stuff
        if self.mode == "paint":
            x = mx
            y = my

            x -= Global.cam.x
            y -= Global.cam.y
            Global.screen.blit(tile, (x, y))
            #set tile or erase
            if pygame.mouse.get_pressed()[0]:
                #check if mouse is on tilemap
                x, y = mx // tilemap.tileDim[0], my // tilemap.tileDim[1]
                if x >= 0 and x < tilemap.mapDim[0] and y >= 0 and y < tilemap.mapDim[1]:
                    tilemap.map[int(y)][int(x)] = self.selectedTile
            elif pygame.mouse.get_pressed()[2]:
                x, y = mx // tilemap.tileDim[0], my // tilemap.tileDim[1]
                if x >= 0 and x < tilemap.mapDim[0] and y >= 0 and y < tilemap.mapDim[1]:
                    tilemap.map[int(y)][int(x)] = 0
        elif self.mode == "select":
            #paint mouse hovered tile
            x, y = Keys.mouse_x, Keys.mouse_y
            x = (x // width)
            y = (y // height)
            outline = pygame.Rect(x * width, y * width, width, height)
            pygame.draw.rect(Global.screen, "black", outline, 2)
            if pygame.mouse.get_pressed()[0]:
                x *= tilemap.orginDim[0]
                y *= tilemap.orginDim[1]
                self.selectedTile = tilemap.posToTile[(int(x), int(y))]

        #move selection by arrow keys
        x, y = tilemap.tilePosInImage[self.selectedTile]

        width = 16
        height = 16

        if Keys.is_pressed(Keys.left): x -= width
        if Keys.is_pressed(Keys.right): x += width
        if Keys.is_pressed(Keys.up): y -= height
        if Keys.is_pressed(Keys.down): y += height

        if (int(x), int(y)) in tilemap.posToTile:
            self.selectedTile = tilemap.posToTile[(int(x), int(y))]

class _collisionsPencil():
    def __init__(self, tilemap):
        self.selectedTile = 0
        self.shapes = tilemap.collisionShapes.num
        self.mode = "paint"
        size = tilemap.tilescale / 2
        self.size = tilemap.orginDim[0] * size
        self.coll = _tileCollisionDefs((tilemap.orginDim[0] * size, tilemap.orginDim[1] * size))

    def update(self, tilemap):
        x, y = Keys.mouse_x, Keys.mouse_y
        #draw tilesets and background
        Global.screen.fill("darkgrey")
        image = pygame.transform.scale_by(tilemap.tileset, tilemap.tilescale / 2)
        Global.screen.blit(image, (0, 0))
        Global.screen.blit(image, (0, image.get_size()[1]+ 10))
        #draw all collision rects on top tileset
        for tile in range(tilemap.tilesetNum - 1):
            tile+=1
            if str(tile) not in tilemap.collisionDict: continue
            rectshape = tilemap.collisionDict[str(tile)]
            rect = getattr(self.coll, rectshape).copy()
            tx, ty = tilemap.tilePosInImage[int(tile)]
            tx = (tx // tilemap.orginDim[0]) * self.size
            ty = (ty // tilemap.orginDim[1]) * self.size
            rect.x += tx
            rect.y += ty
            pygame.draw.rect(Global.screen, "yellow", rect, 5)
            outline = pygame.Rect(tx, ty, self.size, self.size)
            pygame.draw.rect(Global.screen, "gray", outline, 2)
        #change mode
        if y > image.get_size()[1] * 2 + 50: self.mode = "select"
        else: self.mode = "paint"
        #draw collisions tiles at bottom
        tx, ty = 0, image.get_size()[1] * 2 + 50
        for tile in range(self.shapes-1):
            rect = getattr(self.coll, tilemap.collisionShapes.rects[tile]).copy()
            rect.x += tx
            rect.y += ty
            pygame.draw.rect(Global.screen, "yellow", rect, 5)
            outline = pygame.Rect(tx, ty, self.size, self.size)
            pygame.draw.rect(Global.screen, "gray", outline, 2)
            tx += self.size + self.size / 2
        #paint mode
        if self.mode == "paint":
            #draw tile around mouse
            rect = getattr(self.coll, tilemap.collisionShapes.rects[self.selectedTile]).copy()
            rect.x += x - self.size / 2
            rect.y += y - self.size / 2
            pygame.draw.rect(Global.screen, "yellow", rect, 5)
            outline = pygame.Rect(x, y, self.size, self.size)
            outline.center = (Keys.mouse_x, Keys.mouse_y)
            pygame.draw.rect(Global.screen, "gray", outline, 2)
            #paint collisions
            size = tilemap.orginDim[0]
            if pygame.mouse.get_pressed()[0]:
                x = (x - (x % self.size)) // self.size
                y = (y - (y % self.size)) // self.size
                px = x * size
                py = y * size
                if (int(px), int(py)) in tilemap.posToTile:
                    tilemap.collisionDict[str(tilemap.posToTile[(int(px), int(py))])] = self.coll.rects[self.selectedTile]
            #erase collisions
            if pygame.mouse.get_pressed()[2]:
                x = (x - (x % self.size)) // self.size
                y = (y - (y % self.size)) // self.size
                px = x * size
                py = y * size
                if (int(px), int(py)) in tilemap.posToTile:
                    tilemap.collisionDict[tilemap.posToTile[(int(px), int(py))]] = "none"

        elif self.mode == "select":
            #get tile at mouse
            x = int(x // (self.size + self.size / 2))
            #draw underline
            underline = pygame.Rect((0, 0), (self.size, 5))
            underline.center = (Keys.mouse_x, ty + self.size+5)
            pygame.draw.rect(Global.screen, "black", underline)
            #select tile
            if pygame.mouse.get_pressed()[0] and x < self.shapes-1 and x >= 0:
                self.selectedTile = int(x)
                




import pygame
import numpy as np


from .basics._net import Global
from .player._player import Player

class Game:
    def __init__(self, width, height, bg_color, WinTitle="GameBox"):
        """
        Initialize the Game object.

        Parameters:
        width (int): The width of the game window.
        height (int): The height of the game window.
        bg_color (tuple): The background color of the game window.
        WinTitle (str): The title of the game window.

        Returns:
        None
        """
        Global.screenDim = (width, height)
        Global.screen = pygame.display.set_mode(Global.screenDim)
        pygame.display.set_caption(WinTitle)
        Global.bgColor = bg_color
        Global.clock = pygame.time.Clock()
        Global.game = self
        self.objs = []
        self.ui_objs = []


    def update(self, event: pygame.event,frame_rate=60):
        Global.event = event
        #clear collisions
        Global.collisions.clear()

        Global.dt = Global.clock.tick(frame_rate) / 1000.0
        Global.screen.fill(Global.bgColor)
        player = None
        for obj in self.objs:
            if type(obj) == Player: player = obj
            else: obj.update()
        if player != None: player.update()
        
        #update ui
        for obj in self.ui_objs:
            obj.update()
        
        pygame.display.update()

    def get_screen(self):
        return Global.screen
    
    def quit(self):
        #will save files later
        for obj in self.objs:
            if hasattr(obj, "_quit") and callable(obj._quit):
                obj._quit()

    def _set_debug(self, debug: bool):
        """
        Set the debug mode for the game. This will show / hide debuging information 
        like collisions. Note: This feture is for development and testing purposes only.
        
        Parameters:
        debug (bool): The debug mode.
        
        Returns:
        None
        """
        Global._debug = debug
    def _get_debug_state(self):
        """
        Get the debug mode for the game.
        
        Returns:
        bool: The debug mode.
        """
        return Global._debug





import pygame
import numpy as np

from ..basics.utils import zeroOut
from ..basics._net import Global

class _Conditions:
    def __init__(self):
        #conditions start with 'C' and then command ('v' for velocity), and dir (^ up, _ down, < left, > right, # none, ~ any)
        self.velocity_up = 'CV^'
        self.velocity_down = 'CV_'
        self.velocity_left = 'CV<'
        self.velocity_right = 'CV>'
        self.velocity_none = 'CV#'
        self.velocity_any = 'CV~'

Conditions = _Conditions()
Global.cond = Conditions

class Condition_check:
    def __init__(self, stateTree: dict[str, dict[_Conditions, str]], currentState: str):
        self.stateTree = stateTree
        self.currentState = currentState
        
    def check(self, velocity: tuple, pos: tuple):
        state = self.stateTree[self.currentState]
        for cond, next in state.items():
            #velocities
            if cond[1] == 'V':
                if self._resolve_velocities(velocity, cond):
                    self.currentState = next
                    return next
        return self.currentState
                

    def _resolve_velocities(self, velocities, cond):
        vx, vy = velocities
        vx = zeroOut(vx, 0.1)
        vy = zeroOut(vy, 0.1)
        dir = cond[2]
        print(vx, vy, dir)
        #resolve in order up, down, left, right, none, any
        if dir == "^" and vy < 0: return True
        if dir == "_" and vy > 0: return True
        if dir == "<" and vx < 0: return True
        if dir == ">" and vx > 0: return True
        if dir == "#" and vx == 0 and vy == 0: return True
        if dir == "~" and (vx != 0 or vy != 0): return True
        return False

    def updateState(self, state: str):
        self.currentState = state
    def updateStateTree(self, stateTree: dict[str, dict[_Conditions, str]]):
        self.stateTree = stateTree
            




import pygame
import numpy as np

from ..basics._net import Global

from ..GameLevel_ui._sprites import Sprite_2d

class Image:
    def __init__(self, pos: tuple, image, scale: float = 1.0):
        """
        Initialize an Image UI element.
        
        Args:
            pos: Tuple (x, y) for the image position
            image: Either a file path (str) or pygame.Surface object
            scale: Scale factor for the image (default: 1.0)
        """
        #add to game
        self.image = Sprite_2d(pos, image, scale)
        self.image.__remove__()
        Global.game.ui_objs.append(self)

    def move_by(self, x: int, y: int):
        self.image.move_by(x, y)

    def move_to(self, x: int, y: int):
        self.image.move_to(x, y)
    
    def get_pos(self):
        return self.image.get_pos()

    def rescale(self, scale: float):
        self.image.rescale(scale)

    def update(self):
        Global.screen.blit(self.image.image, self.image.pos)



import pygame
import numpy as np

class _global_:
    def __init__(self):
        #screen info and win data
        self.screen: pygame.Surface = None
        self.dt: float = 0
        self.screenDim: tuple = None
        self.bgColor: tuple = None
        self.clock: pygame.time.Clock = None

        #basics
        #--game and cam
        self.game: object = None
        self.cam: object = None
        self.event: pygame.event = None
        #-collisions
        self.collisions: list[pygame.Rect] = []

        self.cond: object = None

        self._debug: bool = False

        #objects
        self.player = self._player()
        self.tilemap = []

    class _player:
        def __init__(self):
            self.pos: tuple = None
            self.player: object = None



Global = _global_()


import pygame

from ..basics._net import Global


class _keys:
    def __init__(self):

        #alphabet
        self.a = pygame.K_a
        self.b = pygame.K_b
        self.c = pygame.K_c
        self.d = pygame.K_d
        self.e = pygame.K_e
        self.f = pygame.K_f
        self.g = pygame.K_g
        self.h = pygame.K_h
        self.i = pygame.K_i
        self.j = pygame.K_j
        self.k = pygame.K_k
        self.l = pygame.K_l
        self.m = pygame.K_m
        self.n = pygame.K_n
        self.o = pygame.K_o
        self.p = pygame.K_p
        self.q = pygame.K_q
        self.r = pygame.K_r
        self.s = pygame.K_s
        self.t = pygame.K_t
        self.u = pygame.K_u
        self.v = pygame.K_v
        self.w = pygame.K_w
        self.x = pygame.K_x
        self.y = pygame.K_y
        self.z = pygame.K_z

        #numbers
        self.n0 = pygame.K_0
        self.n1 = pygame.K_1
        self.n2 = pygame.K_2
        self.n3 = pygame.K_3
        self.n4 = pygame.K_4
        self.n5 = pygame.K_5
        self.n6 = pygame.K_6
        self.n7 = pygame.K_7
        self.n8 = pygame.K_8
        self.n9 = pygame.K_9

        #special keys
        self.enter = pygame.K_RETURN
        self.escape = pygame.K_ESCAPE
        self.space = pygame.K_SPACE
        self.backspace = pygame.K_BACKSPACE
        self.tab = pygame.K_TAB

        #arrow keys
        self.up = pygame.K_UP
        self.down = pygame.K_DOWN
        self.left = pygame.K_LEFT
        self.right = pygame.K_RIGHT

        #mouse
        self.mouse_x, self.mouse_y = 0, 0
    def init(self): Global.game.objs.append(self)

    def is_pressed(self, key): 
        for event in Global.event:
            if event.type == pygame.KEYDOWN:
                if event.key == key: return True
        return False

    def update(self): self.mouse_x, self.mouse_y = pygame.mouse.get_pos()


Keys = _keys()


import pygame
import numpy as np

from ._net import Global

from ..basics.utils import moveTward

class Cammera:
    def __init__(self, scale: float = 1.0):
        self.x = 0
        self.y = 0
        Global.game.objs.append(self)
        Global.cam = self
        self.follow = None
        self.diff = (0, 0)
        self.scale = scale

    def _move(self, x: int, y: int):
        self.x += x
        self.y += y

    def update(self):
        return
        if self.follow is not None:
            self.x = (self.follow.x - self.diff[0])
            self.y = (self.follow.y - self.diff[1])

    def set_follow_target(self, target: object):
        self.follow = target
        print(target)
        self.diff = (target.x - self.x, target.y - self.y)

    def set_scale(self, scale: float = 1.0):
        self.scale = scale
    def change_scale(self, scale: float = 1.0):
        self.scale += scale



import pygame
import numpy as np

from ..basics._net import Global
from ._Animations import Animation

class Sprite_2d:
    def __init__(self, pos: tuple, image, scale: float = 1.0, collision = True, dirrection: int = 1):
        """
        Initialize a 2D sprite.
        
        Args:
            pos: Tuple (x, y) for the sprite position
            image: Either a file path (str) or pygame.Surface object
            scale: Scale factor for the sprite (default: 1.0)
        """
        #add to game
        Global.game.objs.append(self)
        self.collision = collision
        self.__worldPos__ = True
        self.dir = dirrection

        self.pos = pos
        if type(image) == str:
            self.image = pygame.image.load(image)
        else:
            self.image = image
        
        #scale image
        print(self.image)
        self.image = pygame.transform.scale_by(self.image, scale)
        #flip image
        if self.dir == -1:
            self.image = pygame.transform.flip(self.image, True, False)

    def update(self):
        #world space
        x, y = self.pos
        if self.__worldPos__:
            x = x - Global.cam.x
            y = y - Global.cam.y
        Global.screen.blit(self.image, (x, y))
        if self.collision:
            rect = self.image.get_rect()
            rect.x = x
            rect.y = y
            Global.collisions.append(rect)

    def switch_dirrection(self):
        self.image = pygame.transform.flip(self.image, True, False)
        

    def move_by(self, x: int, y: int):
        self.pos = (self.pos[0] + x, self.pos[1] + y)

    def move_to(self, x: int, y: int):
        self.pos = (x, y)

    def get_pos(self):
        return self.pos

    def rescale(self, scale: float):
        self.image = pygame.transform.scale_by(self.image, scale)

    def __remove__(self):
        Global.game.objs.remove(self)

def split_image(image, tileDim, startPos):
    if type(image) == str:
        image = pygame.image.load(image)
    else:
        image = image

    #return image split
    x = startPos[0] * tileDim[0]
    y = startPos[1] * tileDim[1]
    return image.subsurface((x, y, tileDim[0], tileDim[1]))

class Animated_Sprite2D:
    def __init__(self, pos, image, imageDim, tileDim, frames, speed, scale = 1.0, collision = True, dirrection = 1):
        self.animation = Animation(image, tileDim, (0, 0), frames, speed)
        self.collision = collision
        self.dir = dirrection
        self.pos = pos

        self.scale = scale
        self.image = pygame.transform.scale_by(self.animation.getFrame(), scale)
        if self.dir == -1:
            self.image = pygame.transform.flip(self.image, True, False)

        self.__worldPos__ = True

        #add to game
        Global.game.objs.append(self)

    def update(self):
        self.animation.update(Global.dt)
        self.image = pygame.transform.scale_by(self.animation.getFrame(), self.scale)
        if self.dir == -1:
            self.image = pygame.transform.flip(self.image, True, False)
        
        #world space
        x, y = self.pos
        if self.__worldPos__:
            x = x - Global.cam.x
            y = y - Global.cam.y
        Global.screen.blit(self.image, (x, y))
        if self.collision:
            rect = self.image.get_rect()
            rect.x = x
            rect.y = y
            Global.collisions.append(rect)

    def __remove__(self):
        Global.game.objs.remove(self)

    def switch_dirrection(self):
        self.image = pygame.transform.flip(self.image, True, False)

    def move_by(self, x: int, y: int):
        self.pos = (self.pos[0] + x, self.pos[1] + y)

    def move_to(self, x: int, y: int):
        self.pos = (x, y)

    def get_pos(self):
        return self.pos

    def rescale(self, scale: float):
        self.image = pygame.transform.scale_by(self.image, scale)

class AnimationPlayer2D:
    def __init__(self, pos, scale):
        self.pos = pos
        self.scale = scale
        self.anims = {}
        self.currentAnim = None

        self.__worldPos__ = True

        #add to game
        Global.game.objs.append(self)

    def update(self):
        if self.currentAnim is not None:
            self.anims[self.currentAnim].update()

    def add_animation(self, name, image, imageDim, tileDim, frames, speed, scale = 1.0, collision = True, dirrection = 1):
        self.anims[name] = Animated_Sprite2D(self.pos, image, imageDim, tileDim, frames, speed, scale, collision, dirrection)
        self.currentAnim = name
        self.anims[self.currentAnim].__remove__()

    def remove_animation(self, name: str):
        self.anims[name].__remove__()
        del self.anims[name]

    def set_worldPos(self, worldPos: bool):
        for anim in self.anims:
            self.anims[anim].__worldPos__ = worldPos

    def __remove__(self):
        if self in Global.game.objs:
            Global.game.objs.remove(self)\

    def set_scale(self, scale: float):
        for anim in self.anims:
            self.anims[anim].rescale(scale)

    def set_animation(self, anim: str):
        self.currentAnim = anim

    def move_by(self, x: int, y: int):
        self.pos = (self.pos[0] + x, self.pos[1] + y)
        for anim in self.anims:
            self.anims[anim].move_by(x, y)

    def move_to(self, x: int, y: int):
        self.pos = (x, y)
        for anim in self.anims:
            self.anims[anim].move_to(x, y)

    def get_pos(self):
        return self.pos





import pygame
import numpy as np
import json

from ..basics._net import Global

from ._collisionDef import _tileCollisionDefs
from ._Editor import _tilemapEditor

class TileMap:
    def __init__(self, tileSet: str, tileDim: tuple, tileScale: float, mapDim: tuple, mapFill: int, saveFile = None):
        self.tilesetFile = tileSet
        self.mapFile = saveFile
        self.tileDim = (tileDim[0] * tileScale, tileDim[1] * tileScale)
        self.tileDim = (self.tileDim[0] * Global.cam.scale, self.tileDim[1] * Global.cam.scale)
        self.mapDim = mapDim
        self.tilescale = tileScale
        self.orginDim = tileDim

        self.tilesetNum = 0

        self.editor = None

        Global.game.objs.append(self)

        self.collisionShapes = _tileCollisionDefs(self.tileDim)

        self.collisionDict = {}

        self.tilePosInImage = {}
        self.posToTile = {}

        #map, tile splitting, ect
        #--create map
        self.map = np.full(self.mapDim, mapFill)
        #--split map into tiles
        self.tiles = {}
        tileset = pygame.image.load(tileSet).convert_alpha()
        self.tileset = tileset
        tile_w, tile_h = tileDim
        tile_id = 1
        tileset_w, tileset_h = tileset.get_size()

        for y in range(0, tileset_h, tile_h):
            for x in range(0, tileset_w, tile_w):
                tile = pygame.Surface(tileDim, pygame.SRCALPHA)
                tile.blit(tileset, (0, 0), (x, y, tile_w, tile_h))
                self.tiles[tile_id] = pygame.transform.scale(tile, self.tileDim)
                self.tilePosInImage[tile_id] = (x, y)
                self.posToTile[(x, y)] = tile_id
                self.tilesetNum += 1
                tile_id += 1

        Global.tilemap.append(self)

    def load_map_from_json(self, filePath: str):
        with open(filePath, "r") as f:
            data = json.load(f)
        self.__private_loadData(filePath, data)

    def __private_loadData(self, path: str, data: dict):
        self.map = np.array(data["map"])
        self.mapFile = path
        self.collisionDict = data["collisions"]

    def activate_editor(self, activation):
        print(f"editor activated. press {activation} to toggle")
        self.editor = _tilemapEditor(self, activation)

    def update(self):
        self.draw_tiles()
        if self.editor is not None: self.editor._update()
            

    def draw_tiles(self):
        for y in range(self.mapDim[1]):
            for x in range(self.mapDim[0]):
                if self.map[y][x] == 0:
                    continue
                tile = self.tiles[self.map[y][x]]
                mx = (x * self.tileDim[0]) - Global.cam.x
                my = (y * self.tileDim[1]) - Global.cam.y
                if mx < -self.tileDim[0] or mx > Global.screenDim[0] or my < -self.tileDim[1] or my > Global.screenDim[1]: continue
                Global.screen.blit(tile, (mx, my))

    def _quit(self):
        #save map
        if self.mapFile is not None:
            print('tilemap saved')
            with open(self.mapFile, "w") as f:
                data = {}
                data["map"] = self.map.tolist()
                data["collisions"] = self.collisionDict
                json.dump(data, f)



import pygame
import numpy as np

from ..basics._net import Global

from ..player._playerPhysics import _playerPhysics
from ..player._playerSprite import _playerSprite

from ..GameLevel_ui._sprites import Sprite_2d

class Player:
    def __init__(self, pos: tuple, size: tuple, color: tuple = (0, 0, 0), gravity: bool = False):
        self.x, self.y = pos
        self.screenPos = pos
        self.dim = size
        self.color = color
        self.width, self.height = size

        self.gravity = gravity

        self.state = ""
        
        Global.game.objs.append(self)
        Global.player.pos = pos
        Global.player.player = self

        self.sprite = _playerSprite(self)


    def add_physics(self, speed: float = 1.0, gravity: float = 0.0, jump: float = 10.0, maxV: float = 10.0, airRes: float = 0.2):
        self.physics = _playerPhysics(self, speed, gravity, jump, maxV, airRes)

    def update(self):
        self.physics.update()
        #ui

        if (Global.cam.follow) != (self):
            x = self.x - Global.cam.x
            y = self.y - Global.cam.y
        elif (Global.cam.follow) == (self):
            x = self.x
            y = self.y
        velocity = (self.physics.vx, self.physics.vy)
        self.sprite.update(x, y, velocity)

    #movement
    def top_down_movement(self):
        self.physics.top_down_movement()

    def platforming_movement(self):
        self.physics.platforming_movement()

    def set_tilemap_sample(self, sample: int = 10):
        """
        Sets the sample size for player physics collisions. 
        Is the radius of tiles that will be used to get tilemap collisions around player. Note: 
        The larger the sample size the longer it may take to calculate collisions per frame.
        """
        self.physics.sample = sample





import pygame
import numpy as np

from ..basics._net import Global
from ..basics.utils import clamp, moveTward, zeroOut
from ..helpers._collisions import CheckCollisions

class _playerPhysics:
    def __init__(self, player, speed: float = 1.0, gravity: float = 0.0, jump: float = 0.0, maxV: float = 10.0, airRes: float = 0.2):
        self.player = player
        self.rect = pygame.Rect(self.player.x, self.player.y, self.player.dim[0], self.player.dim[1])
        
        self.speed = speed
        self.gravity = gravity
        self.jump = jump
        self.player = player
        self.maxV = maxV
        self.airRes = airRes
        self.onGround = False
        self.canJump = False

        self.sample = 25

        self.vx, self.vy = 0, 0

    def update(self):
        self.ckeck_is_on_ground()
        x, y, self.vx, self.vy = CheckCollisions(self.player.x, self.player.y, 
                               self.vx, self.vy, self.player.dim, 
                               self.sample,self.player)
        if (Global.cam.follow) != (self.player):
            self.player.x = x
            self.player.y = y
        elif (Global.cam.follow) == (self.player):
            dx = -(self.player.x - x)
            dy = -(self.player.y - y)
            Global.cam._move(self.vx, self.vy)



    
    def ckeck_is_on_ground(self):
        self.canJump = False
        rect = pygame.Rect(self.player.x, self.player.y + self.player.dim[1], self.player.dim[0], 10)
        for collision in Global.collisions:
            if collision.colliderect(rect):
                self.canJump = True


    def move(self, x, y):
        self.vx += x * self.speed
        self.vy += y * self.speed

    def top_down_movement(self):
        keys = pygame.key.get_pressed()
        if keys[pygame.K_a]:
            self.move(-1, 0)
        if keys[pygame.K_d]:
            self.move(1, 0)
        if keys[pygame.K_w]:
            self.move(0, -1)
        if keys[pygame.K_s]:
            self.move(0, 1)
        #limit velocity
        self.vx = clamp(self.vx, -self.maxV, self.maxV)
        self.vy = clamp(self.vy, -self.maxV, self.maxV)
        self.vx = moveTward(self.vx, 0, self.airRes)
        self.vy = moveTward(self.vy, 0, self.airRes)
        self.vx = zeroOut(self.vx, 0.3)
        self.vy = zeroOut(self.vy, 0.3)

    def platforming_movement(self):
        keys = pygame.key.get_pressed()
        if keys[pygame.K_a]:
            self.move(-1, 0)
        if keys[pygame.K_d]:
            self.move(1, 0)
        if keys[pygame.K_w] and self.canJump:
            self.vy -= self.jump
        #limit velocity
        self.vx = clamp(self.vx, -self.maxV, self.maxV)
        self.vx = moveTward(self.vx, 0, self.airRes)
        self.vx = zeroOut(self.vx, 0.3)
        if self.player.gravity and not self.onGround: self.vy += self.gravity



import pygame
import numpy as np

from ..basics._net import Global

class Rect:
    def __init__(self, pos: tuple, size: tuple, color: tuple = (0, 0, 0), collision: bool = True):
        self.x, self.y = pos
        self.width, self.height = size
        self.color = color
        Global.game.objs.append(self)
        self.collision = collision

    def update(self):
        width = self.width * Global.cam.scale
        height = self.height * Global.cam.scale
        if (Global.cam.follow) != (self):
            x = self.x - Global.cam.x
            y = self.y - Global.cam.y
        elif (Global.cam.follow) == (self):
            x = self.x
            y = self.y

        rect = pygame.Rect(x, y, width ,height)
        if self.collision: Global.collisions.append(rect)
        pygame.draw.rect(Global.screen, self.color, rect)

    def move(self, x, y):
        if (Global.cam.follow) != (self):
            self.x += x
            self.y += y
        else:
            Global.cam._move(x, y)
            
    def move_to(self, x, y):
        if (Global.cam.follow) != (self):
            self.x = x
            self.y = y
        else:
            Global.cam.x = x
            Global.cam.y = y

    def __remove__(self):
        Global.game.objs.remove(self)



import pygame
import numpy as np

from ..basics._net import Global


def CheckCollisions(x, y, vx, vy, dim, sample, obj):
        x, y = x + vx, y + vy

        #basic object collisions
        x, y, vx, vy = _mainCollisionLogic(Global.collisions, x, y, vx, vy, dim)
        x, y, vx, vy = _checkTilemapCollisions(x, y, vx, vy, dim, sample, obj)

        return x, y, vx, vy
    
def _mainCollisionLogic(collisions, x, y, vx, vy, dim):
    # Y-axis collisions
    py = y
    new_rect = pygame.Rect((x, y), dim)
    if Global._debug:
        pygame.draw.rect(Global.screen, "green", new_rect, 5)
    for collision in collisions:
        if Global._debug:
            pygame.draw.rect(Global.screen, "yellow", collision, 5)
        if collision.colliderect(new_rect):
            if vy > 0:  # falling
                y = collision.top - dim[1]
                vy = 0
            elif vy < 0:  # jumping
                y = collision.bottom
                vy = 0
                
    new_rect = pygame.Rect((x, py), dim)
    for collision in collisions:
        if Global._debug:
            pygame.draw.rect(Global.screen, "yellow", collision, 5)
        if collision.colliderect(new_rect):
            if vx > 0:
                x = collision.left - dim[0]
            elif vx < 0:
                x = collision.right
            vx = 0

    return x, y, vx, vy
    
def _checkTilemapCollisions(x, y, vx, vy, dim, sample, obj):
    if len(Global.tilemap) == 0:
        return x, y, vx, vy
    
    for tilemap in Global.tilemap:
    
        #get player reletive tilemap  pos
        prect = pygame.Rect(x, y, dim[0], dim[1])
        prx, pry = prect.center
        if Global.cam.follow != obj:
            if Global.cam.x != 0 or Global.cam.y != 0:
                px = int((prx) - Global.cam.x)
                py = int((pry) - Global.cam.y) 
                px = int(px  // tilemap.tileDim[0])
                py = int(py  // tilemap.tileDim[1])
            else:
                px = int((x) / tilemap.tileDim[0])
                py = int((y ) / tilemap.tileDim[1])
        else:
            if Global.cam.x != 0 or Global.cam.y != 0:
                px = int(((prx + Global.cam.x)) // tilemap.tileDim[0])
                py = int(((pry + Global.cam.y)) // tilemap.tileDim[1])
            else:
                px = int((x) / tilemap.tileDim[0])
                py = int((y) / tilemap.tileDim[1])


        #check if player is on tilemap
        if px < 0 or px >= tilemap.mapDim[0] or py < 0 or py >= tilemap.mapDim[1]:
            return x, y, vx, vy
        #get collision rects around player
        collisions: list[pygame.Rect] = []
        for tx in range(px - sample, px + sample):
            for ty in range(py - sample, py + sample):
                nx = int(px + tx)
                ny = int(py + ty)
                #if tile is on map
                if nx < 0 or nx >= tilemap.mapDim[0] or ny < 0 or ny >= tilemap.mapDim[1]:
                    continue
                #if tile has defined collision shape
                tile = str(tilemap.map[ny][nx])
                if tile not in tilemap.collisionDict: continue

                #get collision shape 
                rectshape = tilemap.collisionDict[str(tile)]
                rect = getattr(tilemap.collisionShapes, rectshape).copy()

                # Position rect correctly in the world
                if Global.cam.follow != obj:
                    if Global.cam.x != 0 or Global.cam.y != 0:
                        rect.x = ((nx * tilemap.tileDim[0])) - Global.cam.x
                        rect.y = ((ny * tilemap.tileDim[1])) - Global.cam.y
                    else:
                        rect.x += (nx * tilemap.tileDim[0]) - Global.cam.x
                        rect.y += (ny * tilemap.tileDim[1]) - Global.cam.y
                else:
                    if Global.cam.x != 0 or Global.cam.y != 0:
                        rect.x += ((nx * tilemap.tileDim[0])) - Global.cam.x
                        rect.y += ((ny * tilemap.tileDim[1])) - Global.cam.y
                    else:
                        rect.x += (nx * tilemap.tileDim[0]) - Global.cam.x
                        rect.y += (ny * tilemap.tileDim[1]) - Global.cam.y
                collisions.append(rect)

        #check collisions
        x, y, vx, vy = _mainCollisionLogic(collisions, x, y, vx, vy, dim)

        return x, y, vx, vy




import pygame
import numpy as np

from ..basics._net import Global

from ..GameLevel_ui._sprites import Sprite_2d, Animated_Sprite2D, AnimationPlayer2D
from ..helpers._Conditions import Condition_check, _Conditions

class _playerSprite:
    def __init__(self, player):
        self.player = player

        self.state = player.state
        self.statetree = {}
        self.condition_check = None

        self.sprite = None
        

    def update(self, x, y, velocity):
        #state updates
        if self.condition_check is not None:
            self.state = self.condition_check.check(velocity, self.player.screenPos)
            #set animation if possible
            if type(self.sprite) == AnimationPlayer2D:
                if self.state in self.sprite.anims:
                    self.sprite.set_animation(self.state)
        #any changes to follow target
        if Global.cam.follow == self.player and self.sprite is not None:
            self.sprite.__worldPos__ = False
            self.sprite.move_to(self.player.screenPos[0], self.player.screenPos[1])
        elif Global.cam.follow != self.player and self.sprite is not None:
            self.sprite.__worldPos__ = True
            self.sprite.move_to(self.player.x, self.player.y)
        
        if self.sprite is None:
            #rect (x, y) is top left corner
            rect = pygame.Rect((x, y), self.player.dim)
            pygame.draw.rect(Global.screen, self.player.color, rect)
        if type(self.sprite) == Sprite_2d:
            self.sprite.update()
        elif type(self.sprite) == Animated_Sprite2D:
            self.sprite.update()
        elif type(self.sprite) == AnimationPlayer2D:
            self.sprite.update()

    def add_sprite_2d(self, image, scale=1.0, dirrection=1):
        self.sprite = Sprite_2d((self.player.x, self.player.y), image, scale, False, dirrection)
        if Global.cam.follow == self.player:
            self.sprite.__worldPos__ = False
        self.sprite.__remove__()

    def add_animated_sprite_2d(self, image, imageDim, tileDim, frames, speed, scale = 1.0, collision = True, dirrection = 1):
        self.sprite = Animated_Sprite2D((self.player.x, self.player.y), image, imageDim, tileDim, frames, speed, scale, collision, dirrection)
        if Global.cam.follow == self.player:
            self.sprite.__worldPos__ = False
        self.sprite.__remove__()

    def remove_sprite(self):
        if self.sprite is not None:
            self.sprite = None

    def add_animation_player(self, scale = 1.0):
        self.sprite = AnimationPlayer2D((self.player.x, self.player.y), scale)
        if Global.cam.follow == self.player:
            self.sprite.set_worldPos(False)
        self.sprite.__remove__()

    def add_animation(self, name, image, imageDim, tileDim, frames, speed, scale = 1.0, dirrection = 1):
        self.sprite.add_animation(name, image, imageDim, tileDim, frames, speed, scale, False, dirrection)
        self.sprite.__remove__()
        self.sprite.set_worldPos(False)

    def remove_animation(self, name: str):
        self.sprite.remove_animation(name)

    def set_animation(self, anim: str):
        self.sprite.set_animation(anim)

    def set_states(self, stateTree: dict[str, dict[_Conditions, str]], currentState: str):
        self.statetree = stateTree
        self.state = currentState
        self.condition_check = Condition_check(self.statetree, self.state)
        


